package com.sd.business.service.inventory.impl;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.collection.CollectionUtil;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.ruoyi.common.exception.ServiceException;
import com.ruoyi.common.utils.wrapper.IWrapper;
import com.sd.business.entity.bom.po.Bom;
import com.sd.business.entity.bom.po.BomSpec;
import com.sd.business.entity.in.po.InOutStorageDetails;
import com.sd.business.entity.inventory.bo.InOutFun;
import com.sd.business.entity.inventory.dto.InventorySelectDto;
import com.sd.business.entity.inventory.dto.QuantityByWarehouseDto;
import com.sd.business.entity.inventory.po.BalancePriceNum;
import com.sd.business.entity.inventory.po.Inventory;
import com.sd.business.entity.inventory.po.InventoryBackup;
import com.sd.business.entity.inventory.vo.InventoryVo;
import com.sd.business.entity.inventory.vo.QuantityByDepartmentVo;
import com.sd.business.entity.inventory.vo.QuantityByWarehouseVo;
import com.sd.business.mapper.inventory.InventoryMapper;
import com.sd.business.service.bom.BomSpecService;
import com.sd.business.service.in.InOutStorageDetailsService;
import com.sd.business.service.in.impl.InOutStorageServiceImpl;
import com.sd.business.service.inventory.InventoryService;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;


/**
 * <p>
 * 库存 服务实现类
 * </p>
 *
 * @author
 * @since 2023-07-03
 */
@Service
public class InventoryServiceImpl extends ServiceImpl<InventoryMapper, Inventory> implements InventoryService {

    @Autowired
    private BomSpecService bomSpecService;

    @Autowired
    private InOutStorageDetailsService inOutStorageDetailsService;

    @Override
    public Page<InventoryVo> getPage(InventorySelectDto dto, String tableName) {
        IWrapper<Inventory> wrapper = getWrapper();

        wrapper.orderByDesc("i", Inventory::getId);
        wrapper.eq("b", Bom::getBomClassifyId, dto.getBomClassifyId());
        wrapper.like("bs", BomSpec::getName, dto.getBomSpecName());
        wrapper.like("bs", BomSpec::getCode, dto.getBomSpecCode());
        wrapper.eq("b", Bom::getSpecies, dto.getBomSpecies());
        wrapper.eq("b", Bom::getChromatophore, dto.getBomChromatophore());
        wrapper.eq("b", Bom::getEmbossingProcess, dto.getBomEmbossingProcess());
        wrapper.eq("b", Bom::getFrontGrain, dto.getBomFrontGrain());
        wrapper.eq("b", Bom::getReverseGrain, dto.getBomReverseGrain());
        wrapper.like("bs", BomSpec::getColour, dto.getBomSpecColour());
        wrapper.like("bs", BomSpec::getLength, dto.getBomSpecLength());
        wrapper.like("bs", BomSpec::getWidth, dto.getBomSpecWidth());
        wrapper.eq("i", Inventory::getWarehouseId, dto.getWarehouseId());
        wrapper.eq("i", Inventory::getDepartmentId, dto.getDepartmentId());
        wrapper.eq("i", InventoryBackup::getBackupDate, dto.getBackupDate());

        Page<InventoryVo> page = this.baseMapper.getPage(dto.getPage(), wrapper, tableName);
        return page;
    }


    @Override
    public List<QuantityByWarehouseVo> getQuantityByWarehouse(QuantityByWarehouseDto dto) {
        return baseMapper.getQuantityByWarehouse(dto);
    }

    @Override
    public List<QuantityByDepartmentVo> getQuantityByDepartment() {
        return baseMapper.getQuantityByDepartment();
    }

    @Override
    public void in(Long departmentId, Long warehouseId, List<? extends InOutFun> list) {
        List<Long> bomSpecIdList = list.stream().map(InOutFun::getBomSpecId).collect(Collectors.toList());

        synchronized (this) {
            List<Inventory> inventoryList = getInventoryList(departmentId, warehouseId, bomSpecIdList);
            Map<Long, Inventory> map = inventoryList.stream().collect(Collectors.toMap(Inventory::getBomSpecId, Function.identity()));

            for (InOutFun inOutFun : list) {
                Long bomSpecId = inOutFun.getBomSpecId();
                BigDecimal quantity = inOutFun.getQuantity();
                BigDecimal unitPrice = InOutStorageServiceImpl.UNITPRICE;
                Inventory inventory = map.get(bomSpecId);

                if (inventory != null) {
                    inventory.setQuantity(inventory.getQuantity().add(quantity));
                } else {
                    inventory = new Inventory();
                    inventory.setWarehouseId(warehouseId);
                    inventory.setDepartmentId(departmentId);
                    inventory.setBomSpecId(bomSpecId);
                    inventory.setQuantity(quantity);
                    inventoryList.add(inventory);
                    map.put(bomSpecId, inventory);
                }

                List<InOutStorageDetails> detailsList = getDetailsList(warehouseId, bomSpecIdList);
                if (CollUtil.isEmpty(detailsList)){
                    for (InOutStorageDetails details : detailsList) {
                        details.setInSumPrice(BigDecimal.ZERO);
                        details.setQuantity(BigDecimal.ZERO);
                    }
                }
                BigDecimal sumPrice = detailsList.stream().map(InOutStorageDetails::getInSumPrice).reduce(BigDecimal.ZERO, BigDecimal::add);
                BigDecimal sumQuantity = detailsList.stream().map(InOutStorageDetails::getQuantity).reduce(BigDecimal.ZERO, BigDecimal::add);

                // 计算最新的结存数量和结存金额
                sumQuantity = sumQuantity.add(quantity);

                BigDecimal balancep = quantity.multiply(unitPrice);
                sumPrice = sumPrice.add(balancep);
                BigDecimal balanceUnitPrice = sumPrice.divide(sumQuantity, 2, BigDecimal.ROUND_HALF_UP);
                inventory.setBalanceUnitPrice(balanceUnitPrice);
            }

            saveOrUpdateBatch(inventoryList);
        }

    }

    @Override
    public void out(Long departmentId, Long warehouseId, List<? extends InOutFun> list) {
        Map<Long, BigDecimal> map = list.stream().collect(Collectors.toMap(InOutFun::getBomSpecId, InOutFun::getQuantity));
        List<Long> bomSpecIdList = new ArrayList<>(map.keySet());

        synchronized (this) {
            List<Inventory> inventoryList = getInventoryList(departmentId, warehouseId, bomSpecIdList);

            // 查询库存中是否存在bom
            for (Inventory inventory : inventoryList) {
                Long bomSpecId = inventory.getBomSpecId();
                if (!bomSpecIdList.contains(bomSpecId)) {
                    BomSpec bomSpec = bomSpecService.getById(bomSpecId);
                    if (bomSpec == null) {
                        throw new ServiceException("未知bom规格id：" + bomSpecId);
                    }
                    throw new ServiceException("出库失败,品名为：" + bomSpec.getName() + "的bom库存不足");
                }
            }

            //根据仓库id和规格id获取出入详细表数据
            List<InOutStorageDetails> detailsList = getDetailsList(warehouseId, bomSpecIdList);
            if (CollUtil.isEmpty(detailsList)){
                throw new ServiceException("数据不存在");
            }
            //存储需要修改的入库记录
            List<InOutStorageDetails> updateList = new ArrayList<>();
            //存储这次出库的金额
            Map<Long, BigDecimal> outPriceMap = new HashMap<>();

            int j = 0;
            BigDecimal num;
            BigDecimal quantity;
            do {
                //获取这条记录剩余的数量
                InOutStorageDetails details = detailsList.get(j);

                //获取这次一种规格的出库的数量
                quantity = map.get(details.getBomSpecId());

                //获取两次数量直接的差价
                num = details.getQuantity().subtract(quantity);

                //如果这条记录库存不足,则计算下一条记录
                if (num.compareTo(BigDecimal.ZERO) < 0){
                    details.setQuantity(BigDecimal.ZERO);
                    j++;
                    continue;
                }
                //对数量进行减扣
                //BigDecimal detailsNum = details.getQuantity().subtract(num);
                details.setQuantity(num.abs());

                //数量减少了,那金额也得减少
                details.setInSumPrice(num.abs().multiply(details.getInUnitPrice()));

                //将这次出库的金额存储到map里面
                BigDecimal outPrice = details.getInUnitPrice().multiply(num);
                outPriceMap.put(details.getBomSpecId(), outPrice);
                //将减扣数量后的对象修改回去
                updateList.add(details);
                j++;

            }while (num.compareTo(BigDecimal.ZERO) < 0 && j < detailsList.size() );

                inOutStorageDetailsService.updateBatchById(updateList);


                for (Inventory inventory : inventoryList) {
                    // 获取之前的结存金额
                    BigDecimal balancePrice = inventory.getQuantity().multiply(inventory.getBalanceUnitPrice());
                    // 减去出库金额后的结存金额
                    BigDecimal newBalancePrice;
                    if (outPriceMap.get(inventory.getBomSpecId()) == null){
                        newBalancePrice = balancePrice.subtract(BigDecimal.ZERO);
                    }else {
                        newBalancePrice = balancePrice.subtract(outPriceMap.get(inventory.getBomSpecId()));
                    }
                    // 减去出库数量后的结存数量
                    inventory.setQuantity(inventory.getQuantity().subtract(map.get(inventory.getBomSpecId())));
                    // 获取最后的结存单价
                    BigDecimal balanceUnitPrice = newBalancePrice.divide(inventory.getQuantity(), 2, BigDecimal.ROUND_HALF_UP);
                    //将数据存入对象
                    inventory.setBalanceUnitPrice(balanceUnitPrice);
                }
                updateBatchById(inventoryList);
            }

    }
        private List<Inventory> getInventoryList (Long departmentId, Long warehouseId, List < Long > bomSpecIdList){
            return list(q -> q
                    .eq(Inventory::getDepartmentId, departmentId)
                    .eq(Inventory::getWarehouseId, warehouseId)
                    .in(Inventory::getBomSpecId, bomSpecIdList)
            );
        }

        private List<InOutStorageDetails> getDetailsList (Long warehouseId, List < Long > bomSpecIdList){
            return inOutStorageDetailsService.list(q -> q
                    .eq(InOutStorageDetails::getWarehouseId, warehouseId)
                    .in(InOutStorageDetails::getBomSpecId, bomSpecIdList)
                    .isNotNull(InOutStorageDetails::getQuantity)

            );
        }

}
